from __future__ import annotations
from gi.repository import Gio, GLib, Gtk

from typing import Callable, Dict, Optional, Set, Any

try:
    from importlib.resources import read_text
except ModuleNotFoundError:
    from importlib_resources import read_text  # type: ignore


ActionCallback = Optional[Callable[[Gio.SimpleAction, Any], None]]


def print_event(*args: Any, **kwargs: Any) -> None:
    """
    Print out all arguments for event exploration, usage: gtk_widget.connect(gtk_event_name, print_event)
    """
    def nice(e: Any) -> Any:
        if isinstance(e, Gtk.Widget):
            return e.get_path().to_string()
        else:
            return e

    print(list(map(nice, args)), kwargs)


class ActionRegistry:
    def __init__(self, for_widget: Optional[Gio.ActionMap] = None):
        self.for_widget = for_widget if for_widget else Gio.SimpleActionGroup()
        self.actions: Dict[str, Gio.SimpleAction] = {}

    def create(self, name: str, callback: ActionCallback = None,
               param_type: GLib.VariantType = None, state: GLib.Variant = None) -> Gio.SimpleAction:
        callback = callback if callback else getattr(self.for_widget, 'on_' + name)
        action = Gio.SimpleAction(name=name, parameter_type=param_type, state=state)
        if state is not None:
            action.connect("change-state", callback)
        else:
            action.connect("activate", callback)
        self.for_widget.add_action(action)
        self.actions[name] = action
        return action

    def __getitem__(self, item: str) -> Gio.SimpleAction:
        return self.actions[item]


class Callback:
    """
    Wrapper around a callback function with equality by callback function, to be used by WhenIdle
    """
    def __init__(self, callback: Callable, *args: Any, **kwargs: Any):  # type: ignore[type-arg]
        self.priority = kwargs.pop('priority', 100)
        self.callback = callback
        self.args = args
        self.kwargs = kwargs

    def __hash__(self) -> int:
        return hash(self.callback)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Callback) and self.callback == other.callback

    def __call__(self) -> None:
        self.callback(*self.args, **self.kwargs)

    def __str__(self) -> str:
        return '{:03d} {!r}({!r}, {!r})'.format(self.priority, self.callback, self.args, self.kwargs)


class WhenIdle:
    """
    Debouncing wrapper around GLib.idle_add (or another call-later-mechanism) that overwrites prior calls to the same callback

    Also supports priorities, lower priorities get called first

    Usage: see WhenIdle.call
    """
    _instance: WhenIdle = None

    def __init__(self, runner_callback: Callable):  # type: ignore[type-arg]
        self._runner_callback = runner_callback
        self._callbacks: Set[Callback] = set()

    @classmethod
    def instance(cls) -> WhenIdle:
        if cls._instance is None:
            cls._instance = cls(GLib.idle_add)
        return cls._instance

    @classmethod
    def call(cls, callback: Callable, *args: Any, **kwargs: Any) -> None:   # type: ignore[type-arg]
        """
        Debouncing wrapper around GLib.idle_add
        Usage:
        > WhenIdle.call(self.redraw, 5, force = True, priority=10)
        will call self.redraw(5, force=True) with priority 10
        """
        cls.instance().add(Callback(callback, *args, **kwargs))

    def add(self, callback: Callback) -> None:
        self._callbacks.add(callback)
        self._runner_callback(self._run)

    def _run(self) -> None:
        if self._callbacks:
            callback = sorted(self._callbacks, key=lambda c: int(c.priority))[0]
            callback()
            self._callbacks.remove(callback)
            self._runner_callback(self._run)


def resource_string(resource: str, package: str = 'ocrd_browser.resources') -> str:
    return read_text(package, resource)



from gi.repository import Gio, Gtk, GLib, Gdk

from typing import List, Dict, Any

from ocrd_browser import __version__
from ocrd_browser.util.gtk import ActionRegistry
from ocrd_browser.ui import MainWindow, WindowFlags, AboutDialog, OpenDialog
from ocrd_browser.view import ViewRegistry

from importlib_metadata import entry_points


class OcrdBrowserApplication(Gtk.Application):
    # TODO: Parse arguments (with Gtk or click) to open certain views by mets+page_id+view(view_configuration_dict) and deep filename e.g.
    # OCR-D-OCR-TESS-frk-SEG-LINE-tesseract-ocropy-DEWARP/OCR-D-OCR-TESS-frk-SEG-LINE-tesseract-ocropy-DEWARP_0001.xml ->
    # file_grp => OCR-D-OCR-TESS-frk-SEG-LINE-tesseract-ocropy-DEWARP /     <mets:fileGrp USE="OCR-D-OCR-TESS-frk-SEG-LINE-tesseract-ocropy-DEWARP"><mets:file MIMETYPE="application/vnd.prima.page+xml" ID="OCR-D-OCR-TESS-frk-SEG-LINE-tesseract-ocropy-DEWARP_0001"><mets:FLocat LOCTYPE="OTHER" OTHERLOCTYPE="FILE" xlink:href="OCR-D-OCR-TESS-frk-SEG-LINE-tesseract-ocropy-DEWARP/OCR-D-OCR-TESS-frk-SEG-LINE-tesseract-ocropy-DEWARP_0001.xml"/></mets:file>
    # mimetype => application/vnd.prima.page+xml
    # FILEID => OCR-D-OCR-TESS-frk-SEG-LINE-tesseract-ocropy-DEWARP_0001 /  <mets:fptr FILEID="OCR-D-OCR-TESS-frk-SEG-LINE-tesseract-ocropy-DEWARP_0001"/>
    # page.id => PHYS_0017 / <mets:div TYPE="page" ID="PHYS_0017">
    # mets.xml + page.id + view_by_mimetype(mimetype)(file_grp = (file_grp,mimetype))
    # TODO: Test application startup for all kinds of mets.xml (not necessarily OCR-D mets.xml) and views without crashes
    def __init__(self) -> None:
        Gtk.Application.__init__(self, application_id='org.readmachine.ocrd-browser',
                                 flags=Gio.ApplicationFlags.HANDLES_OPEN)
        self.actions = ActionRegistry(for_widget=self)
        self.view_registry = ViewRegistry.create_from_entry_points()

        self.window_flags = WindowFlags.NONE
        self.restricted = False

        self.add_main_option("version", ord("v"), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, "Show version and exit", None)
        self.add_main_option("maximize", ord("m"), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, "Open in maximized window", None)
        self.add_main_option("fullscreen", ord("f"), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, "Open in fullscreen window", None)
        self.add_main_option("restricted", ord("r"), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, "Open restricted (no edit/open functionality)", None)

    def do_startup(self) -> None:
        Gtk.Application.do_startup(self)
        self.actions.create('new')
        self.actions.create('open')
        self.actions.create('about')
        self.actions.create('quit')

        self.set_accels_for_action('view.zoom_by(1)', ['<Ctrl>plus'])
        self.set_accels_for_action('view.zoom_by(-1)', ['<Ctrl>minus'])
        self.set_accels_for_action('view.zoom_to::original', ['<Ctrl>0'])
        self.set_accels_for_action('view.zoom_to::width', ['<Ctrl>numbersign'])
        self.set_accels_for_action('view.zoom_to::page', ['<Ctrl><Alt>numbersign'])
        self.set_accels_for_action("win.fullscreen", ['F11'])

        for entry_point in entry_points(group='ocrd_browser_ext'):
            (entry_point.load())(self)

        self.load_css()

    def load_css(self) -> None:
        css = Gtk.CssProvider()
        css.load_from_resource('/org/readmachine/ocrd-browser/css/theme.css')
        # noinspection PyArgumentList
        Gtk.StyleContext().add_provider_for_screen(Gdk.Screen.get_default(), css, Gtk.STYLE_PROVIDER_PRIORITY_USER)
        # css = Gtk.CssProvider()
        # css.load_from_path('/home/jk/PycharmProjects/ocrd-browser/gresources/css/test.css')
        # Gtk.StyleContext().add_provider_for_screen(Gdk.Screen.get_default(), css, Gtk.STYLE_PROVIDER_PRIORITY_USER)

    def do_activate(self) -> None:
        self._present_window(self.get_active_window())

    def on_about(self, _action: Gio.SimpleAction, _param: str = None) -> None:
        about_dialog = AboutDialog(application=self, transient_for=self.get_active_window(), modal=True)
        about_dialog.present()

    def on_quit(self, _action: Gio.SimpleAction, _param: str = None) -> None:
        open_windows: int = 0
        window: MainWindow
        for window in self.get_windows():
            if isinstance(window, MainWindow) and window.close_confirm():
                window.destroy()
            else:
                open_windows += 1
        if open_windows == 0:
            self.quit()

    def on_open(self, _action: Gio.SimpleAction, _param: str = None) -> None:
        open_dialog = OpenDialog(application=self, transient_for=self.get_active_window(), modal=True)
        response = open_dialog.run()
        if response == Gtk.ResponseType.OK:
            self.open_in_window(open_dialog.get_uri(), window=open_dialog.get_transient_for())
        open_dialog.destroy()

    def on_new(self, _action: Gio.SimpleAction, _param: str = None) -> None:
        win = MainWindow(application=self, restricted=self.restricted)
        win.present()

    def do_open(self, files: List[Gio.File], file_count: int, hint: str) -> int:
        for file in files:
            self.open_in_window(file.get_uri(), window=None)
        return 0

    def do_handle_local_options(self, goptions: GLib.VariantDict) -> int:
        options: Dict[str, Any] = goptions.end().unpack()
        if options.get('version', False):
            print('browse-ocrd: {}'.format(__version__))
            return 0
        if options.get('maximize', False):
            self.window_flags |= WindowFlags.MAXIMIZE
        if options.get('fullscreen', False):
            self.window_flags |= WindowFlags.FULLSCREEN
        self.restricted = options.get('restricted', False)

        return -1

    def open_in_window(self, uri: str, window: MainWindow = None) -> None:
        if not window or not window.document.empty:
            window = MainWindow(application=self, restricted=self.restricted)
        self._present_window(window)
        GLib.timeout_add(10, window.open, uri)

    def _present_window(self, win: MainWindow) -> None:
        if not win:
            win = MainWindow(application=self, restricted=self.restricted)
        if self.window_flags & WindowFlags.MAXIMIZE:
            win.maximize()
        if self.window_flags & WindowFlags.FULLSCREEN:
            win.fullscreen()

        win.present()



from gi.repository import Gtk, GLib, GdkPixbuf

from typing import Tuple, Optional, Dict, List, Union, NewType, Callable, Any
from itertools import count

from ocrd_browser.util.image import cv_to_pixbuf, cv_scale
from ocrd_browser.model import Document
from .icon_store import LazyLoadingListStore
from ..util.config import SettingsFactory

import cv2
import os

RowResult = Tuple[Optional[int], Optional[Gtk.TreeModelRow]]
ChangeList = Union[List[str], Dict[str, str]]
Column = NewType('Column', int)


class PageListStore(LazyLoadingListStore):
    """
    PageListStore is a GTK.ListStore for use with GTK.IconView and works as an adapter to ocrd_browser.model.Document

    It utilizes LazyLoadingListStore for lazy thumbnail generation and
    contains the domain specific logic for handling Document events
    """
    COLUMN_PAGE_ID = Column(0)
    COLUMN_TOOLTIP = Column(1)
    COLUMN_FILENAME = Column(2)
    COLUMN_THUMB = Column(3)
    COLUMN_ORDER = Column(4)
    COLUMN_HASH = Column(5)

    def __init__(self, document: Document):
        """
        Initializes the underlying ListStore and fills it with a row for each page, then start the lazy loading

        The actual image and data loading happens in _load_row
        """
        columns = {
            self.COLUMN_PAGE_ID: str,
            self.COLUMN_TOOLTIP: str,
            self.COLUMN_FILENAME: str,
            self.COLUMN_THUMB: GdkPixbuf.Pixbuf,
            self.COLUMN_ORDER: int
            # self.COLUMN_HASH: str file hash = filename + modified_time (gets added by LazyLoadingListStore)
        }
        super().__init__(*(columns.values()), init_row=self._init_row, load_row=self._load_row, hash_row=self._hash_row)
        self.document = document
        self.pixbufs: Dict[str, GdkPixbuf.Pixbuf] = {
            icon_name: GdkPixbuf.Pixbuf.new_from_resource(
                '/org/readmachine/ocrd-browser/icons/{}.png'.format(icon_name)
            ) for icon_name in ['page-loading', 'page-missing']
        }

        # TODO: make file_group selectable, see https://github.com/hnesk/browse-ocrd/issues/7#issuecomment-707851109
        self.file_group = document.get_default_image_group(SettingsFactory.settings().file_groups.preferred_images)
        file_lookup = document.get_image_paths(self.file_group)
        order = count(start=1)
        for page_id in self.document.page_ids:
            file = file_lookup[page_id]
            self.append((page_id, '', str(file) if file else None, None, next(order)))

        GLib.timeout_add(10, self.start_loading)

    def get_row_by_page_id(self, page_id: str) -> RowResult:
        """
        Find index and row by page_id
        """
        return self.get_row_by_column_value(self.COLUMN_PAGE_ID, page_id)

    def get_row_by_column_value(self, column: int, value: str) -> RowResult:
        """
        Find index and row by column value
        """
        for n, row in enumerate(self):
            if row[column] == value:
                return n, row
        return None, None

    def iter_for_id(self, page_id: str) -> Optional[Gtk.TreeIter]:
        """
        Get a Gtk.TreeIter for the page_id
        """
        path = self.path_for_id(page_id)
        return self.get_iter(path) if path else None

    def path_for_id(self, page_id: str) -> Optional[Gtk.TreePath]:
        """
        Get a Gtk.TreePath for the page_id
        """
        n, row = self.get_row_by_page_id(page_id)
        return Gtk.TreePath(n) if n is not None else None

    def document_changed(self, subtype: str, changes: ChangeList) -> None:
        """
        Event callback to sync Document modifications with the ListStore

        @param subtype: str one of  'page_added', 'page_deleted', 'page_changed', 'reordered'
        @param changes: List[str] affected page_ids
        """

        def _page_added(page_ids: List[str]) -> None:
            for page_id in page_ids:
                try:
                    file = next(iter(self.document.workspace.mets.find_files(pageId=page_id, fileGrp=self.file_group.group, mimetype=self.file_group.mime)))
                    file_name = str(self.document.path(file))
                    self.append((page_id, '', file_name, None, len(self)))
                except StopIteration as e:
                    raise ValueError('Page {} in group {}  not in workspace'.format(page_id, self.file_group)) from e

        def _page_deleted(page_ids: List[str]) -> None:
            for delete_page_id in reversed(page_ids):
                n, row = self.get_row_by_page_id(delete_page_id)
                self.remove(self.get_iter(Gtk.TreePath(n)))

        def _page_changed(page_ids: List[str]) -> None:
            for page_id in page_ids:
                n, row = self.get_row_by_page_id(page_id)
                try:
                    file = next(iter(self.document.workspace.mets.find_files(pageId=page_id, fileGrp=self.file_group.group, mimetype=self.file_group.mime)))
                    file_name = str(self.document.path(file))
                    row[self.COLUMN_FILENAME] = file_name
                except StopIteration:
                    pass

        def _reordered(old_to_new_ids: Dict[str, str]) -> None:
            id_to_position: Dict[str, int] = {}
            for n, row in enumerate(self):
                id_to_position[row[self.COLUMN_PAGE_ID]] = n

            positions: List[int] = list(range(0, len(old_to_new_ids)))
            for old, new in old_to_new_ids.items():
                positions[id_to_position[old]] = id_to_position[new]

            self.reorder(positions)

            # Update the order in the ListStore data, not needed for now, but might help if we have sorting
            order = count(start=1)
            for page_id in self.document.page_ids:
                n, row = self.get_row_by_page_id(page_id)
                row[self.COLUMN_ORDER] = next(order)

        handler: Dict[str, Callable[[Any], None]] = {
            'page_added': _page_added,
            'page_deleted': _page_deleted,
            'page_changed': _page_changed,
            'reordered': _reordered,
        }
        handler[subtype](changes)

    def _init_row(self, row: Gtk.TreeModelRow) -> None:
        if row[self.COLUMN_FILENAME] is not None:
            row[1] = 'Loading {}'.format(row[self.COLUMN_FILENAME])
            row[3] = self.pixbufs['page-loading']
        else:
            row[1] = 'No image for {}'.format(row[self.COLUMN_PAGE_ID])
            row[3] = self.pixbufs['page-missing']

    @staticmethod
    def _load_row(row: Gtk.TreeModelRow) -> Gtk.TreeModelRow:
        filename = row[PageListStore.COLUMN_FILENAME]
        if filename is not None:
            image = cv2.imread(filename)
            row[1] = '{} ({}x{})'.format(filename, image.shape[1], image.shape[0])
            row[3] = cv_to_pixbuf(cv_scale(image, 100, None))
        return row

    @staticmethod
    def _hash_row(row: Gtk.TreeModelRow) -> str:
        file = row[PageListStore.COLUMN_FILENAME]
        if file is not None:
            modified_time = os.path.getmtime(file)
            return '{}:{}'.format(file, modified_time)
        else:
            return ''



from .document import Document  # noqa E402
from .page import Page, IMAGE_FROM_PAGE_FILENAME_SUPPORT  # noqa E402
from .page_xml_renderer import PageXmlRenderer

__all__ = ['Page', 'IMAGE_FROM_PAGE_FILENAME_SUPPORT', 'Document', 'PageXmlRenderer']



from gi.repository import Gtk, GdkPixbuf
from typing import Any

from ocrd_browser.util.gtk import resource_string
from ocrd_browser import __version__
from ocrd_browser.model import Document


@Gtk.Template(string=resource_string('about-dialog.ui'))
class AboutDialog(Gtk.AboutDialog):
    __gtype_name__ = "AboutDialog"

    def __init__(self, **kwargs: Any):
        Gtk.AboutDialog.__init__(self, **kwargs)
        # noinspection PyCallByClass,PyArgumentList
        self.set_logo(GdkPixbuf.Pixbuf.new_from_resource('/org/readmachine/ocrd-browser/icons/logo.png'))
        self.set_version(__version__)


@Gtk.Template(string=resource_string('open-dialog.ui'))
class OpenDialog(Gtk.FileChooserDialog):
    __gtype_name__ = "OpenDialog"

    def __init__(self, **kwargs: Any):
        # noinspection PyCallByClass
        Gtk.FileChooserDialog.__init__(self, **kwargs)

        filter_mets = Gtk.FileFilter()
        filter_mets.set_name("METS files")
        filter_mets.add_mime_type("application/mets+xml")
        filter_mets.add_pattern("*.xml")
        self.add_filter(filter_mets)

        filter_any = Gtk.FileFilter()
        filter_any.set_name("Any files")
        filter_any.add_pattern("*")
        self.add_filter(filter_any)


@Gtk.Template(string=resource_string('save-dialog.ui'))
class SaveDialog(Gtk.FileChooserDialog):
    __gtype_name__ = "SaveDialog"

    def __init__(self, **kwargs: Any):
        # noinspection PyCallByClass
        Gtk.FileChooserDialog.__init__(self, **kwargs)

        filter_text = Gtk.FileFilter()
        filter_text.set_name("METS files")
        filter_text.add_mime_type("text/xml")
        self.add_filter(filter_text)

        filter_any = Gtk.FileFilter()
        filter_any.set_name("Any files")
        filter_any.add_pattern("*")
        self.add_filter(filter_any)


class SaveChangesDialog(Gtk.MessageDialog):

    def __init__(self, document: Document, **kwargs: Any):
        Gtk.MessageDialog.__init__(self, **kwargs)
        name = document.workspace.mets.unique_identifier if document.workspace.mets.unique_identifier else '<unnamed>'
        self.props.text = 'Save changes to document "{}" before closing?'.format(name)
        self.props.secondary_text = 'If you don’t save, changes to {} will be permanently lost.'.format(
            document.original_url if document.original_url else 'this file')
        self.props.message_type = Gtk.MessageType.QUESTION

        # Close Button
        close_button: Gtk.Button = self.add_button('Close without saving', Gtk.ResponseType.NO)
        close_button.get_style_context().add_class('destructive-action')

        # Cancel Button
        self.add_button('_Cancel', Gtk.ResponseType.CANCEL)

        # Save Button
        save_text = '_Save' if document.original_url else '_Save as ...'
        save_button: Gtk.Button = self.add_button(save_text, Gtk.ResponseType.YES)

        self.set_default(save_button)



from gi.repository import GObject, GtkSource, Gtk, Gdk, Pango

from typing import Optional, Any

from ocrd_utils.constants import MIMETYPE_PAGE

from ocrd_browser.util.file_groups import FileGroupHandle
from ocrd_browser.view import View
from ocrd_browser.view.base import FileGroupSelector, FileGroupFilter

GObject.type_register(GtkSource.View)


class ViewText(View):
    """
    A view of the current PAGE-XML TextEquiv annotation concatenated as plain text
    """

    label = 'Text'

    def __init__(self, name: str, window: Gtk.Window):
        super().__init__(name, window)
        self.file_group = FileGroupHandle(None, MIMETYPE_PAGE)
        self.font_size: Optional[int] = None
        # noinspection PyTypeChecker
        self.text_view: GtkSource.View = None
        # noinspection PyTypeChecker
        self.buffer: GtkSource.Buffer = None

    def build(self) -> None:
        super().build()
        self.add_configurator('file_group', FileGroupSelector(FileGroupFilter.PAGE))

        self.text_view = GtkSource.View(visible=True, vexpand=False, editable=False,
                                        monospace=False, show_line_numbers=True, width_request=400)
        self.buffer = self.text_view.get_buffer()
        self.text_view.connect('scroll-event', self.on_scroll)
        self.scroller.add(self.text_view)

    @property
    def use_file_group(self) -> str:
        return self.file_group.group

    def config_changed(self, name: str, value: Any) -> None:
        super().config_changed(name, value)
        self.reload()

    def on_scroll(self, _widget: GtkSource.View, event: Gdk.EventScroll) -> bool:
        # Handles zoom in / zoom out on Ctrl+mouse wheel
        accel_mask = Gtk.accelerator_get_default_mod_mask()
        if event.state & accel_mask == Gdk.ModifierType.CONTROL_MASK:
            did_scroll, delta_x, delta_y = event.get_scroll_deltas()
            if did_scroll and abs(delta_y) > 0:
                self.zoom(delta_y)
                return True
        return False

    def zoom(self, direction: float = 0.0) -> None:
        """
        Zoom in or out by direction
        :param direction: amount to zoom

        TODO: make it DRY, currently copy-pasted in ViewText, ViewXml and ViewDiff (maybe as a hidden Configurator)
        """
        style: Gtk.StyleContext = self.text_view.get_style_context()
        font: Pango.FontDescription = style.get_font(style.get_state())
        if self.font_size is None:
            self.font_size = font.get_size()
        size = self.font_size
        size *= (1.2 ** direction)
        if 1 * Pango.SCALE > size or size > 100 * Pango.SCALE:
            return
        self.font_size = int(size)
        font.set_size(size)
        self.text_view.override_font(font)

    def redraw(self) -> None:
        if self.current:
            self.text_view.set_tooltip_text(self.page_id)
            regions = self.current.pc_gts.get_Page().get_AllRegions(classes=['Text'], order='reading-order')
            text = ''
            for i, region in enumerate(regions):
                if i:
                    text += '\n'  # or line feed?
                if region.get_TextEquiv() and region.get_TextEquiv()[0].Unicode:
                    text += region.get_TextEquiv()[0].Unicode
                    continue
                lines = region.get_TextLine()
                for j, line in enumerate(lines):
                    if j:
                        text += '\n'
                    if line.get_TextEquiv() and line.get_TextEquiv()[0].Unicode:
                        text += line.get_TextEquiv()[0].Unicode
                        continue
                    words = line.get_Word()
                    for k, word in enumerate(words):
                        if k:
                            text += ' '
                        if word.get_TextEquiv() and word.get_TextEquiv()[0].Unicode:
                            text += word.get_TextEquiv()[0].Unicode
                            continue
                        glyphs = word.get_Glyph()
                        for l, glyph in enumerate(glyphs): # noqa E741
                            if glyph.get_TextEquiv() and glyph.get_TextEquiv()[0].Unicode:
                                text += glyph.get_TextEquiv()[0].Unicode

            self.buffer.set_text(text)
        else:
            self.buffer.set_text('')



from typing import Dict, Optional

from gi.repository import Gtk

from . import ViewEmpty
from .base import View
from ..model import Document


class ViewManager:

    def __init__(self, window: Gtk.ApplicationWindow, root: Gtk.Box):
        self.window = window
        self.root = root
        self.views: Dict[str, View] = {}

    def set_root_view(self, view_class: type) -> View:
        view = self._create_view(view_class)
        self.connect(self.window, view)
        self.root.add(view.container)
        self.print()
        return view

    def __getitem__(self, view_name: str) -> View:
        if view_name not in self.views:
            raise ValueError('Unknown view: ' + str(view_name))
        return self.views[view_name]

    def close(self, view_name: str) -> None:
        view = self[view_name]
        paned = view.container.get_parent()
        others = [sibling for sibling in paned.get_children() if sibling != view.container]

        if paned is self.root:
            raise ValueError('Not removing root')
        if not isinstance(paned, Gtk.Paned):
            raise ValueError('parent is no Paned but ' + str(type(paned)))
        if len(others) != 1:
            raise ValueError('more than 1 sibling')

        other = others[0]
        parent: Gtk.Container = paned.get_parent()
        paned.remove(view.container)
        paned.remove(other)
        view.container.destroy()
        self.disconnect(self.window, view)

        parent.remove(paned)
        parent.add(other)

        del self.views[view_name]
        del view
        self.print()

    def add(self, new_view_type: Optional[type] = None) -> View:
        return self.split(None, new_view_type, False)

    def split(self, split_view_name: Optional[str] = None, new_view_type: Optional[type] = None, vertical: bool = True) -> View:
        split_view_name = split_view_name or list(self.views.keys())[0]
        new_view_type = new_view_type or ViewEmpty
        view = self[split_view_name]
        parent: Gtk.Box = view.container.get_parent()
        new_pane_root = Gtk.Paned(visible=True, orientation=Gtk.Orientation(vertical))
        new_view = self._create_view(new_view_type)
        parent.remove(view.container)
        new_pane_root.pack1(view.container, True, False)
        new_pane_root.pack2(new_view.container, True, False)
        parent.add(new_pane_root)
        self.connect(self.window, new_view)
        self.print()
        return new_view

    def replace(self, replace_view: str, new_view_type: type) -> View:
        view = self[replace_view]
        parent: Gtk.Box = view.container.get_parent()
        new_view = self._create_view(new_view_type, replace_view)
        parent.remove(view.container)
        parent.add(new_view.container)
        self.views[view.name] = new_view
        view.container.destroy()
        self.connect(self.window, new_view)
        self.print()
        return new_view

    def _create_view(self, view_class: type, name: Optional[str] = None) -> View:
        name = name or self._unique_name()
        view: View = view_class(name, self.window)
        view.build()
        view.set_document(self.window.document)
        view.page_activated(self.window, self.window.current_page_id)
        self.views[name] = view
        return view

    def set_document(self, document: Document) -> None:
        for view in self.views.values():
            view.set_document(document)

    def update_ui(self) -> None:
        for view in self.views.values():
            view.update_ui()

    def _unique_name(self) -> str:
        for i in range(len(self.views.keys()), 999999):
            name = 'view_' + str(i)
            if name not in self.views:
                return name
        raise RuntimeError('Found no unique view name')

    def disconnect(self, win: Gtk.Window, view: View) -> None:
        try:
            win.disconnect_by_func(view.page_activated)
            win.disconnect_by_func(view.pages_selected)
        except Exception as e:
            print(e)

    def connect(self, win: Gtk.Window, view: View) -> None:
        win.connect('page_activated', view.page_activated)
        win.connect('pages_selected', view.pages_selected)

    def print(self) -> None:
        return
        # print(self.views)
        # print(self.print_level())

    def print_level(self, root: Gtk.Paned = None, indent: int = 0) -> str:
        current: Gtk.Container = root or self.root
        s: str = ('\t' * indent) + current.get_name()
        if isinstance(current, Gtk.Box) and current.get_name().startswith('view_'):
            view = self[current.get_name()]
            s = s + ': ' + type(view).__qualname__ + "\n"
        else:
            s = s + "\n"
            for c in current.get_children():
                if isinstance(c, Gtk.Container):
                    s = s + self.print_level(c, indent + 1)
        return s



from __future__ import annotations

import os
import re
import shlex

from collections import defaultdict
from configparser import ConfigParser
from gi.repository import GLib
from ocrd_utils import getLogger
from pydantic import BaseSettings, BaseModel as PydanticBaseModel, Field, validator, Extra
from pydantic.env_settings import SettingsSourceCallable
from shutil import which
from typing import List, Optional, Dict, Any, Tuple


class _DummyObject:
    """
    Dummy object to test formatstrings with nested properties
    """

    def __init__(self, **kwargs: Any):
        for k, v in kwargs.items():
            setattr(self, k, shlex.quote(v) if isinstance(v, str) else v)


class BaseModel(PydanticBaseModel):
    class Config:
        env_nested_delimiter = '__'
        extra = Extra.forbid


def snake_case(optionstr: str) -> str:
    return re.sub('(?!^)([A-Z]+)', r'_\1', optionstr).lower()


def _split_regexes(cls: Any, v: str) -> List[re.Pattern[str]]:
    """
    Split a string at ',' and returns the parts as compiled regexes
    """
    try:
        return [re.compile(r.strip()) for r in v.split(',')]
    except re.error as e:
        raise ValueError(f'Error in pattern "{e.pattern!r}" @ pos {e.pos}: "{e.msg}"') from e


def _check_commandline(cls: Any, v: str, **placeholders: _DummyObject) -> str:
    """
    Checks for a valid shell command string with placeholders

    1) does the executable exists
    2) Are only existing placeholders used
    """
    executable, *rest = shlex.split(v)
    if not which(executable):
        raise ValueError(f'Command "{executable}" not found: "{v}"')
    try:
        v.format(**placeholders)
    except Exception as e:
        raise ValueError(f'{e!s} in "{v}"') from e
    return v


DUMMY_FILE = _DummyObject(
    ID='ID1',
    basename='IMG1.tif',
    basename_without_extension='IMG1',
    extension='tif',
    fileGrp='IMG',
    loctype='URL',
    mimetype='image/tiff',
    otherloctype=None,
    pageId='PAGE1',
    url='IMG/IMG1.tif',
    path=_DummyObject(absolute='/tmp/test/IMG/IMG1.tif', relative='IMG/IMG1.tif')
)

DUMMY_WORKSPACE = _DummyObject(directory='/tmp/test/', baseurl='/tmp/test/mets.xml')


class FileGroups(BaseModel):
    preferred_images: List[re.Pattern]  # type: ignore[type-arg]

    split_preferred_images = validator('preferred_images', pre=True, allow_reuse=True)(_split_regexes)


class Tool(BaseModel):
    commandline: str
    shortcut: Optional[str]
    name: Optional[str]

    def named(self, name: str) -> Tool:
        self.name = name
        return self

    @validator('commandline')
    def check_commandline(cls, v: str) -> str:
        return _check_commandline(cls, v, file=DUMMY_FILE, workspace=DUMMY_WORKSPACE)


class Settings(BaseSettings):
    file_groups: FileGroups = FileGroups(preferred_images='OCR-D-IMG,OCR-D-IMG.*')
    tool: Dict[str, Tool] = Field({})

    @validator('tool')
    def check_tool(cls, tools: Dict[str, Tool]) -> Dict[str, Tool]:
        return {k.lower(): v.named(k) for k, v in tools.items()}

    class Config:
        env_nested_delimiter = '__'
        extra = Extra.forbid
        env_prefix = 'BROCRD__'

        @classmethod
        def customise_sources(
            cls,
            init_settings: SettingsSourceCallable,
            env_settings: SettingsSourceCallable,
            file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            """Prioritize ENV over .conf files"""
            return env_settings, init_settings, file_secret_settings


class SettingsFactory:
    _settings = None

    @classmethod
    def settings(cls) -> Settings:
        if cls._settings is None:
            cls._settings = cls.build_from_files(cls.get_default_files())
        return cls._settings

    @classmethod
    def get_default_files(cls) -> List[str]:
        config_dirs = GLib.get_system_config_dirs() + [GLib.get_user_config_dir()]
        try:
            config_dirs.append(os.getcwd())
        except FileNotFoundError:
            # ignore deleted cwd in unit tests
            pass

        return [dir_ + '/ocrd-browser.conf' for dir_ in config_dirs]

    @classmethod
    def build_from_files(cls, files: List[str]) -> Settings:
        log = getLogger('ocrd_browser.util.config.SettingsFactory.build_from_files')
        config = ConfigParser()
        setattr(config, 'optionxform', snake_case)
        read_files = config.read(files)
        log.info('Read %d config file(s): %s, tried %s', len(read_files), ', '.join(read_files),
                 ', '.join(str(file) for file in files))
        settings = Settings(**cls.config_to_dict(config))
        print(settings)
        return settings

    @staticmethod
    def config_to_dict(config: ConfigParser) -> Dict[str, Any]:
        d: Dict[str, Any] = defaultdict(dict)
        for section, values in config.items():
            if section == 'DEFAULT':
                continue
            parts = section.split(' ', 2)
            top = snake_case(parts[0])
            sub = parts[1] if len(parts) > 1 else None
            if sub:
                d[top][sub] = dict(values)
            else:
                d[top] = dict(values)

        return dict(d)



from __future__ import annotations

import re
import shlex
import sys

from configparser import ConfigParser
from collections import defaultdict
from pydantic import BaseSettings, BaseModel as PydanticBaseModel, Field, validator
from rich import print as pprint
from shutil import which
from typing import Dict, Any, List, Optional


class _DummyObject:
    """
    Dummy object to test formatstrings with nested properties
    """
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, shlex.quote(v) if isinstance(v, str) else v)


def _snake_case(s: str) -> str:
    return re.sub('(?!^)([A-Z]+)', r'_\1', s).lower()


def _split_regexes(cls, v: str) -> List[re.Pattern]:
    """
    Split a string at ',' and returns the parts as compiled regexes
    """
    try:
        return [re.compile(r.strip()) for r in v.split(',')]
    except re.error as e:
        raise ValueError(f'Error in pattern "{e.pattern}" @ pos {e.pos}: "{e.msg}"') from e


def _check_commandline(cls, v: str, **placeholders: _DummyObject) -> str:
    """
    Checks for a valid shell command string with placeholders

    1) does the executable exists
    2) Are only existing placeholders used
    """
    executable, *rest = shlex.split(v)
    if not which(executable):
        raise ValueError(f'Executable "{executable}" not found in "{v}"')
    try:
        v.format(**placeholders)
    except Exception as e:
        raise ValueError(f'{e!s} in "{v}"') from e
    return v


DUMMY_FILE = _DummyObject(
    ID='ID1',
    basename='IMG1.tif',
    basename_without_extension='IMG1',
    extension='tif',
    fileGrp='IMG',
    loctype='URL',
    mimetype='image/tiff',
    otherloctype=None,
    pageId='PAGE1',
    url='IMG/IMG1.tif',
    path=_DummyObject(absolute='/tmp/test/IMG/IMG1.tif', relative='IMG/IMG1.tif')
)

DUMMY_WORKSPACE = _DummyObject(directory='/tmp/test/', baseurl='/tmp/test/mets.xml')


class BaseModel(PydanticBaseModel):
    class Config:
        env_nested_delimiter = '__'


class Tool(BaseModel):
    commandline: str
    shortcut: Optional[str]

    @validator('commandline')
    @classmethod
    def check_commandline(cls, v):
        return _check_commandline(cls, v, file=DUMMY_FILE, workspace=DUMMY_WORKSPACE)


class View(BaseModel):
    groups: List[re.Pattern]

    split_groups = validator('groups', pre=True, allow_reuse=True)(_split_regexes)


class FileGroups(BaseModel):
    preferred_images: List[re.Pattern]

    split_preferred_images = validator('preferred_images', pre=True, allow_reuse=True)(_split_regexes)


class Settings(BaseSettings):
    file_groups: FileGroups
    tool: Dict[str, Tool] = Field({})
    view: Dict[str, View] = Field({})


configstr = """
[FileGroups]
preferredImages = OCR-D-IMG, OCR-D-IMG-*, ORIGINAL 

[Tool PageViewer]
commandline = /usr/bin/java -jar /home/jk/bin/JPageViewer/JPageViewer.jar --resolve-dir {workspace.directory} {file.path.absolute}

[Tool Open]
commandline = xdg-open {file.path.absolute}
shortcut = o
"""


def convert_config(conf: ConfigParser) -> Dict[str, Any]:
    config: Dict[str, Any] = defaultdict(dict)
    for section, values in conf.items():
        if section == 'DEFAULT':
            continue
        parts = section.split(' ', 2)
        top = _snake_case(parts[0])
        sub = parts[1] if len(parts) > 1 else None
        if sub:
            config[top][sub] = dict(values)
        else:
            config[top] = dict(values)

    return dict(config)


# field.field_info.extra['env_names']
if __name__ == "__main__":
    conf = ConfigParser()
    conf.optionxform = _snake_case
    conf.read_string(configstr)
    config = convert_config(conf)
    pprint(config)
    s = Settings.parse_obj(config)
    pprint(s)
    sys.exit()

# print(dict(config.items()))



from __future__ import annotations
import re
from collections import Counter
from typing import List, Optional, Counter as CounterType, NamedTuple, cast, Tuple, Union, Pattern, Sequence, Any

from ocrd_models import OcrdFile

AnyPattern = Union[Pattern[Any], str]
PatternList = Optional[Sequence[AnyPattern]]


class FileGroupHandle(NamedTuple):
    group: str
    mime: str

    @property
    def key(self) -> str:
        return '|'.join(self)

    def match(self, file: OcrdFile) -> bool:
        return cast(bool, file.fileGrp == self.group and file.mimetype == self.mime)

    def __str__(self) -> str:
        return ' - '.join(self)

    @classmethod
    def cast(cls, tpl: Union[FileGroupHandle, Tuple[str, str]]) -> FileGroupHandle:
        if isinstance(tpl, cls):
            return tpl
        else:
            return cls._make(tpl)


def weight_match(s: str, preferred: PatternList = None) -> float:
    """
    Weights how good a string matches a list of regular expressions
    """
    weight = 0.0
    if preferred:
        ln = float(len(preferred))
        for i, pref in enumerate(preferred):
            if re.fullmatch(pref, s):
                # prefer matches earlier in the list
                weight += (ln - i) / ln
                # break or no break???
    return weight


def best_file_group(file_group_handles: List[FileGroupHandle], preferred_groups: PatternList = None, preferred_mimetypes: PatternList = None, cutoff: float = -9999) -> Optional[FileGroupHandle]:
    file_groups_counter: CounterType[FileGroupHandle] = Counter()
    for file_group_handle in file_group_handles:
        file_groups_counter[file_group_handle] += weight_match(file_group_handle.group, preferred_groups)  # type: ignore[assignment]
        file_groups_counter[file_group_handle] += weight_match(file_group_handle.mime, preferred_mimetypes)  # type: ignore[assignment]
        file_groups_counter[file_group_handle] -= len(file_group_handle.group) * 0.0001  # type: ignore[assignment]
    if file_groups_counter:
        [(file_group_handle, score)] = file_groups_counter.most_common(1)
        if score > cutoff:
            return file_group_handle
        else:
            return None
    else:
        return None



from gi.repository import GObject, Gtk, WebKit2

from typing import Any

from ocrd_browser.util.file_groups import FileGroupHandle
from ocrd_browser.view import View
from ocrd_browser.view.base import FileGroupSelector, FileGroupFilter

GObject.type_register(WebKit2.WebView)


class ViewHtml(View):
    """
    A view of the HTML+CSS annotation (as produced by ocrd-dinglehopper reports).
    """

    label = 'HTML'

    def __init__(self, name: str, window: Gtk.Window):
        super().__init__(name, window)
        self.file_group = FileGroupHandle(None, 'text/html')
        # noinspection PyTypeChecker
        self.web_view: WebKit2.WebView = None

    def build(self) -> None:
        super().build()
        self.add_configurator('file_group', FileGroupSelector(FileGroupFilter.HTML))

        self.web_view = WebKit2.WebView()

        self.scroller.add(self.web_view)

    @property
    def use_file_group(self) -> str:
        return self.file_group.group

    def config_changed(self, name: str, value: Any) -> None:
        super().config_changed(name, value)
        self.reload()

    def redraw(self) -> None:
        if not self.current:
            return

        html_files = self.current.get_files('text/html')
        if not html_files:
            return

        self.web_view.set_tooltip_text(self.page_id)
        self.web_view.load_uri('file://' + str(self.document.path(html_files[0].local_filename)))
        self.web_view.show()



from difflib import SequenceMatcher

from gi.repository import GObject, GtkSource, Gtk, Gdk, Pango

from typing import Optional, Any, NamedTuple, List

from ocrd_models.ocrd_page_generateds import PcGtsType
from ocrd_utils.constants import MIMETYPE_PAGE

from ocrd_browser.model import Page
from ocrd_browser.util.file_groups import FileGroupHandle
from ocrd_browser.view import View
from ocrd_browser.view.base import FileGroupSelector, FileGroupFilter

GObject.type_register(GtkSource.View)


class TaggedText:
    def __init__(self) -> None:
        self.parts: List[TaggedString] = []
        self.len = 0

    def append(self, string: str, tag: Optional[str] = None) -> None:
        self.parts.append(TaggedString(string, self.len, self.len + len(string), tag))
        self.len += len(string)

    def __str__(self) -> str:
        return ''.join(part.text for part in self.parts)


class TaggedString(NamedTuple):
    text: str
    start: int
    end: float
    tag: Optional[str] = None


def diff_strings(text1: str, text2: str) -> TaggedText:
    def isjunk(x: str) -> bool:
        return x in " \t"

    sm = SequenceMatcher(isjunk=isjunk, a=text1, b=text2, autojunk=False)
    parts = TaggedText()
    try:
        if sm.quick_ratio() < 0.5 and len(text1) > 5:
            parts.append('ERROR: Texts differ too much (similarity {:2%}) to compute an actual alignment and comparison.'.format(sm.ratio()), 'deleted')
        else:
            for op, s1, e1, s2, e2 in sm.get_opcodes():
                t1, t2 = text1[s1:e1], text2[s2:e2]
                if op == 'equal':
                    parts.append(t1, None)
                else:
                    if op in ('delete', 'replace'):
                        parts.append(t1, 'deleted')
                    if op in ('insert', 'replace'):
                        parts.append(t2, 'inserted')
    except Exception as err:
        parts.append(str(err), 'deleted')

    return parts


class IdTag(Gtk.TextTag):
    def __init__(self, _id: Optional[str] = None, **properties: Any) -> None:
        super().__init__()
        self.id = _id
        for prop, value in properties.items():
            setattr(self.props, prop, value)


class ViewDiff(View):
    """
    A view of the current PAGE-XML TextEquiv annotation concatenated as plain text
    """

    label = 'Text-Diff'

    def __init__(self, name: str, window: Gtk.Window):
        super().__init__(name, window)
        self.file_group = FileGroupHandle(None, MIMETYPE_PAGE)
        self.file_group2 = FileGroupHandle(None, MIMETYPE_PAGE)
        self.font_size: Optional[int] = None

        self.current2: Optional[Page] = None

        # noinspection PyTypeChecker
        self.text_view: GtkSource.View = None
        # noinspection PyTypeChecker
        self.buffer: GtkSource.Buffer = None

    def build(self) -> None:
        super().build()
        self.add_configurator('file_group', FileGroupSelector(FileGroupFilter.PAGE))
        self.add_configurator('file_group2', FileGroupSelector(FileGroupFilter.PAGE))

        self.text_view = GtkSource.View(visible=True, vexpand=False, editable=False,
                                        monospace=False, show_line_numbers=True, width_request=400)
        self.buffer = self.text_view.get_buffer()
        self.text_view.connect('scroll-event', self.on_scroll)
        self.scroller.add(self.text_view)

    @property
    def use_file_group(self) -> str:
        return self.file_group.group

    def reload(self) -> None:
        self.current2 = self.document.page_for_id(self.page_id, self.file_group2.group)
        super().reload()

    def config_changed(self, name: str, value: Any) -> None:
        super().config_changed(name, value)
        self.reload()

    def on_scroll(self, _widget: GtkSource.View, event: Gdk.EventScroll) -> bool:
        # Handles zoom in / zoom out on Ctrl+mouse wheel
        accel_mask = Gtk.accelerator_get_default_mod_mask()
        if event.state & accel_mask == Gdk.ModifierType.CONTROL_MASK:
            did_scroll, delta_x, delta_y = event.get_scroll_deltas()
            if did_scroll and abs(delta_y) > 0:
                self.zoom(delta_y)
                return True
        return False

    def zoom(self, direction: float = 0.0) -> None:
        """
        Zoom in or out
        :param direction: amount to zoom

        TODO: make it DRY, currently copy-pasted in ViewText, ViewXml and ViewDiff (maybe as a hidden Configurator)
        """
        style: Gtk.StyleContext = self.text_view.get_style_context()
        font: Pango.FontDescription = style.get_font(style.get_state())
        if self.font_size is None:
            self.font_size = font.get_size()
        size = self.font_size
        size *= (1.2 ** direction)
        if 1 * Pango.SCALE > size or size > 100 * Pango.SCALE:
            return
        self.font_size = int(size)
        font.set_size(size)
        self.text_view.override_font(font)

    def redraw(self) -> None:
        if self.current:
            self.text_view.set_tooltip_text(self.page_id)
            text = self.get_page_text(self.current.pc_gts)
            if self.current2:
                text2 = self.get_page_text(self.current2.pc_gts)
                diffed = diff_strings(text, text2)
                self.buffer.set_text(str(diffed))

                for part in diffed.parts:
                    if part.tag:
                        if part.tag == 'deleted':
                            background = 'red'
                        elif part.tag == 'inserted':
                            background = 'green'
                        else:
                            # This shouldn't happen, but when:
                            background = 'yellow'

                        tag = IdTag(None, background=background)
                        self.buffer.get_tag_table().add(tag)

                        self.buffer.apply_tag(
                            tag,
                            self.buffer.get_iter_at_offset(part.start),
                            self.buffer.get_iter_at_offset(part.end)
                        )
            else:
                self.buffer.set_text(text)
        else:
            self.buffer.set_text('')

    def get_page_text(self, pc_gts: PcGtsType) -> str:
        regions = pc_gts.get_Page().get_AllRegions(classes=['Text'], order='reading-order')
        text = ''
        for i, region in enumerate(regions):
            if i:
                text += '\n'  # or line feed?
            if region.get_TextEquiv() and region.get_TextEquiv()[0].Unicode:
                text += region.get_TextEquiv()[0].Unicode
                continue
            lines = region.get_TextLine()
            for j, line in enumerate(lines):
                if j:
                    text += '\n'
                if line.get_TextEquiv() and line.get_TextEquiv()[0].Unicode:
                    text += line.get_TextEquiv()[0].Unicode
                    continue
                words = line.get_Word()
                for k, word in enumerate(words):
                    if k:
                        text += ' '
                    if word.get_TextEquiv() and word.get_TextEquiv()[0].Unicode:
                        text += word.get_TextEquiv()[0].Unicode
                        continue
                    glyphs = word.get_Glyph()
                    for l, glyph in enumerate(glyphs):  # noqa E741
                        if glyph.get_TextEquiv() and glyph.get_TextEquiv()[0].Unicode:
                            text += glyph.get_TextEquiv()[0].Unicode

        return text



from gi.repository import GObject, GtkSource, Gtk, Gdk, Pango

from typing import Optional, Any

from ocrd_utils.constants import MIMETYPE_PAGE
from ocrd_models.ocrd_page import to_xml

from ocrd_browser.util.file_groups import FileGroupHandle
from ocrd_browser.util.launcher import Launcher
from ocrd_browser.view import View
from ocrd_browser.view.base import FileGroupSelector, FileGroupFilter

GObject.type_register(GtkSource.View)


class ViewXml(View):
    """
    A view of the current PAGE-XML with syntax highlighting
    """

    label = 'PAGE-XML'

    def __init__(self, name: str, window: Gtk.Window):
        super().__init__(name, window)
        self.file_group = FileGroupHandle(None, MIMETYPE_PAGE)
        self.font_size: Optional[int] = None
        # noinspection PyTypeChecker
        self.text_view: GtkSource.View = None
        # noinspection PyTypeChecker
        self.buffer: GtkSource.Buffer = None

    def build(self) -> None:
        super().build()
        self.add_configurator('file_group', FileGroupSelector(FileGroupFilter.XML))
        button = Gtk.Button.new_with_label('PageViewer')
        button.connect('clicked', self.open_jpageviewer)
        button.set_visible(True)
        self.action_bar.pack_start(button)

        lang_manager = GtkSource.LanguageManager()
        style_manager = GtkSource.StyleSchemeManager()

        self.text_view = GtkSource.View(visible=True, vexpand=False, editable=False, monospace=True,
                                        show_line_numbers=True, width_request=400)
        self.buffer = self.text_view.get_buffer()
        self.buffer.set_language(lang_manager.get_language('xml'))
        self.buffer.set_style_scheme(style_manager.get_scheme('tango'))
        self.text_view.connect('scroll-event', self.on_scroll)
        self.scroller.add(self.text_view)

    @property
    def use_file_group(self) -> str:
        return self.file_group.group

    def config_changed(self, name: str, value: Any) -> None:
        super().config_changed(name, value)
        self.reload()

    def open_jpageviewer(self, _button: Gtk.Button) -> None:
        if self.current and self.current.page_file:
            Launcher().launch('pageviewer', self.document, self.current.page_file)

    def on_scroll(self, _widget: GtkSource.View, event: Gdk.EventScroll) -> bool:
        # Handles zoom in / zoom out on Ctrl+mouse wheel
        accel_mask = Gtk.accelerator_get_default_mod_mask()
        if event.state & accel_mask == Gdk.ModifierType.CONTROL_MASK:
            did_scroll, delta_x, delta_y = event.get_scroll_deltas()
            if did_scroll and abs(delta_y) > 0:
                self.zoom(delta_y)
                return True
        return False

    def zoom(self, direction: float = 0.0) -> None:
        """
        Zoom in or out by direction
        :param direction: amount to zoom

        TODO: make it DRY, currently copy-pasted in ViewText, ViewXml and ViewDiff (maybe as a hidden Configurator)
        """
        style: Gtk.StyleContext = self.text_view.get_style_context()
        font: Pango.FontDescription = style.get_font(style.get_state())
        if self.font_size is None:
            self.font_size = font.get_size()
        size = self.font_size
        size *= (1.2 ** direction)
        if 1 * Pango.SCALE > size or size > 100 * Pango.SCALE:
            return
        self.font_size = int(size)
        font.set_size(size)
        self.text_view.override_font(font)

    def redraw(self) -> None:
        if self.current:
            self.text_view.set_tooltip_text(self.page_id)
            if self.current.page_file:
                with self.document.path(self.current.page_file).open('r') as f:
                    text = f.read()
            else:
                text = to_xml(self.current.pc_gts)
            self.buffer.set_text(text)
        else:
            self.buffer.set_text('')



from __future__ import annotations
from typing import Dict, Tuple, Optional, Type
from .base import View

from importlib_metadata import entry_points

ViewInfo = Tuple[type, str, str]


class ViewRegistry:
    def __init__(self, views: Dict[str, ViewInfo]):
        self.views = views

    @classmethod
    def create_from_entry_points(cls) -> ViewRegistry:
        views = {}
        # also loads view plugins, e.g. from browse-ocrd-physical-import
        for entry_point in entry_points(group='ocrd_browser_view'):
            view_class = entry_point.load()
            assert issubclass(view_class, View)
            label = view_class.label if hasattr(view_class, 'label') else view_class.__name__
            description = view_class.__doc__.strip()
            views[entry_point.name] = (view_class, label, description)
        return cls(views)

    def get_view_options(self) -> Dict[str, str]:
        return {id_: label for id_, (view_class, label, description) in self.views.items()}

    def get_view(self, id_: str) -> Optional[Type[View]]:
        return self.views[id_][0] if id_ in self.views else None



from gi.repository import Gtk, Gdk, Pango, Gio, GObject

from typing import List, Callable, Optional, Any, cast

from ocrd_browser.util.gtk import resource_string
from ocrd_browser.model import Document
from .page_store import PageListStore, ChangeList


@Gtk.Template(string=resource_string('page-list.ui'))
class PagePreviewList(Gtk.IconView):
    __gtype_name__ = "PagePreviewList"

    def __init__(self, document: Document, **kwargs: Any):
        super().__init__(**kwargs)
        self.context_menu: Gtk.Menu = Gtk.Menu()
        self.setup_ui()
        self.setup_context_menu()
        self.document: Document = document
        self.model: PageListStore = PageListStore(self.document)
        self.set_model(self.model)
        # noinspection PyTypeChecker
        self.current: Optional[Gtk.TreeIter] = None

    def setup_context_menu(self) -> None:
        menu = Gio.Menu()
        action_menu = Gio.Menu()
        action_menu.append('Remove', 'win.page_remove')
        menu.append_section(None, action_menu)
        prop_menu = Gio.Menu()
        prop_menu.append('Properties', 'win.page_properties')
        menu.append_section(None, prop_menu)
        self.context_menu.bind_model(menu, None, True)
        self.context_menu.attach_to_widget(self)
        self.context_menu.show_all()

    def set_document(self, document: Document) -> None:
        self.document = document
        self.model = PageListStore(self.document)
        self.set_model(self.model)

    def document_changed(self, subtype: str, page_ids: ChangeList) -> None:
        self.model.document_changed(subtype, page_ids)
        if subtype == 'page_added' and page_ids:
            self.scroll_to_id(cast(List[str], page_ids)[-1])

    def setup_ui(self) -> None:
        self.set_text_column(PageListStore.COLUMN_PAGE_ID)
        self.set_tooltip_column(PageListStore.COLUMN_TOOLTIP)
        self.set_pixbuf_column(PageListStore.COLUMN_THUMB)
        text_renderers = [cell for cell in self.get_cells() if isinstance(cell, Gtk.CellRendererText)]
        text_renderer: Gtk.CellRendererText = text_renderers[0]
        text_renderer.props.ellipsize = Pango.EllipsizeMode.MIDDLE
        self.connect('button-press-event', self.button_pressed)

    def button_pressed(self, _sender: Gtk.Widget, event: Gdk.EventButton) -> None:
        if event.button == Gdk.BUTTON_SECONDARY and event.type == Gdk.EventType.BUTTON_PRESS:
            result = self.get_item_at_pos(*event.get_coords())
            if result:
                self.emit('on_context_menu', event, result[0], result[1])

    @GObject.Signal(arg_types=[object, object, object])
    def on_context_menu(self, event: Gdk.EventButton, path: Gtk.TreePath, _renderer: Gtk.CellRenderer) -> None:
        if len(self.get_selected_items()) <= 1:
            self.set_cursor(path, None, False)
            self.emit('select_cursor_item')
        self.context_menu.popup_at_pointer(event)

    def get_selected_ids(self) -> List[str]:
        return [self.model[path][PageListStore.COLUMN_PAGE_ID] for path in self.get_selected_items()]

    def goto_index(self, index: int) -> None:
        index = index if index >= 0 else len(self.model) + index
        if 0 <= index < len(self.model):
            self.goto_path(Gtk.TreePath(index))

    def scroll_to_id(self, page_id: str) -> None:
        path = self.model.path_for_id(page_id)
        if path:
            self.scroll_to_path(path, False, 0, 1.0)

    def skip(self, pos: int) -> None:
        if not self.current:
            self.current = self.model.get_iter(Gtk.TreePath(0))
        iterate: Callable[[Gtk.TreeIter], Gtk.TreeIter] = self.model.iter_previous if pos < 0 else self.model.iter_next
        n = self.current
        for _ in range(0, abs(pos)):
            n = iterate(n)
        self.goto_path(self.model.get_path(n))

    def goto_path(self, path: Gtk.TreePath) -> None:
        self.set_cursor(path, None, False)
        self.emit('select_cursor_item')
        self.emit('activate_cursor_item')
        self.grab_focus()

    def do_selection_changed(self) -> None:
        self.emit('pages_selected', self.get_selected_ids())

    def do_item_activated(self, path: Gtk.TreePath) -> None:
        self.current = self.model.get_iter(path)
        self.emit('page_activated', self.model[path][PageListStore.COLUMN_PAGE_ID])

    @GObject.Signal()
    def page_activated(self, page_id: str) -> None:
        pass

    @GObject.Signal(arg_types=[object])
    def pages_selected(self, page_ids: List[str]) -> None:
        pass



import cv2
import struct
import zlib

from typing import Tuple, Union, Any, cast
from PIL.Image import Image, fromarray
from numpy import (
    array as np_array,
    uint8 as np_uint8,
    dtype as np_dtype,
    bool_ as np_bool,
    stack as np_stack,
)
from gi.repository import GdkPixbuf, GLib

try:
    from numpy.typing import NDArray
    numpy_array = NDArray[Any]
except ImportError:
    from numpy import ndarray as numpy_array

__all__ = ['cv_scale', 'cv_to_pixbuf', 'pil_to_pixbuf', 'pil_scale', 'add_dpi_to_png_buffer']


def cv_to_pixbuf(z: numpy_array) -> GdkPixbuf.Pixbuf:
    if z.dtype == np_bool:
        z = (z * 255).astype(np_uint8)
    if z.ndim == 2:
        z = np_stack((z, z, z), axis=2)
    assert z.ndim == 3
    h, w, c = z.shape
    assert c == 3 or c == 4
    if c == 3:
        z = cv2.cvtColor(z, cv2.COLOR_BGR2RGB)
    else:
        z = cv2.cvtColor(z, cv2.COLOR_BGRA2RGBA)
    args = {'colorspace': GdkPixbuf.Colorspace.RGB, 'has_alpha': c == 4, 'bits_per_sample': 8, 'width': w, 'height': h}
    rowstride = w * c
    pb = GdkPixbuf.Pixbuf.new_from_bytes(data=GLib.Bytes(z.tobytes()), rowstride=rowstride, **args)
    return pb


def pil_to_pixbuf(im: Image) -> GdkPixbuf.Pixbuf:
    if im.mode == 'LA':
        bgra = cv2.cvtColor(np_array(im.convert('L'), dtype=np_uint8), cv2.COLOR_GRAY2BGRA)
        bgra[:, :, 3] = np_array(im.getchannel('A'), dtype=np_uint8)
        im = bgra
    elif im.mode == 'RGBA':
        bgra = cv2.cvtColor(np_array(im.convert('RGB'), dtype=np_uint8), cv2.COLOR_RGB2BGRA)
        bgra[:, :, 3] = np_array(im.getchannel('A'), dtype=np_uint8)
        im = bgra
    else:
        im = cv2.cvtColor(np_array(im.convert('RGB'), dtype=np_uint8), cv2.COLOR_RGB2BGR)
    return cv_to_pixbuf(im)


def cv_scale(orig: numpy_array, w: int = None, h: int = None) -> numpy_array:
    """
    Scale a cv2 image
    :param orig: ndarray Original cv2 image
    :param w: New width
    :param h: New height
    :return: ndarray
    """
    height, width, depth = orig.shape
    new_width, new_height = _calculate_scale(width, height, w, h)
    return cast(numpy_array, cv2.resize(orig, (new_width, new_height)))


def pil_scale(orig: Image, w: int = None, h: int = None) -> Image:
    """
    Scale a Pillow image
    :param orig: ndarray Original cv2 image
    :param w: New width
    :param h: New height
    :return: ndarray
    """
    new_width, new_height = _calculate_scale(orig.width, orig.height, w, h)
    # thumb = orig.copy()
    # thumb.thumbnail((new_width, new_height))
    # also allows enlarging:
    if orig.mode.startswith('I'):
        # workaround for Pillow#4402:
        arr = np_array(orig)
        if arr.dtype.kind == 'i':
            # signed integer is *not* trustworthy in this context
            # (usually a mistake in the array interface)
            arr = arr.astype(np_dtype('u' + arr.dtype.name), copy=False)
        if arr.dtype.kind == 'u':
            # integer needs to be scaled linearly to 8 bit
            # of course, an image might actually have some lower range
            # (e.g. 10-bit in I;16 or 20-bit in I or 4-bit in L),
            # but that would be guessing anyway, so here don't
            # make assumptions on _scale_, just reduce _precision_
            arr = arr >> 8 * (arr.dtype.itemsize - 1)
            arr = arr.astype(np_uint8)
        elif arr.dtype.kind == 'f':
            # float needs to be scaled from [0,1.0] to [0,255]
            arr *= 255
            arr = arr.astype(np_uint8)
        orig = fromarray(arr)
    thumb = orig.resize((new_width, new_height))
    return thumb


def add_dpi_to_png_buffer(image_bytes: bytes, dpi: Union[int, Tuple[int, int]] = 300) -> bytes:
    """
    adds dpi information to a png image

    see https://stackoverflow.com/questions/57553641/how-to-save-dpi-info-in-py-opencv/57555123#57555123

    """
    if isinstance(dpi, (int, float)):
        dpi = (dpi, dpi)

    # Find start of IDAT chunk
    idat_offset = image_bytes.find(b'IDAT') - 4

    # Create our lovely new pHYs chunk - https://www.w3.org/TR/2003/REC-PNG-20031110/#11pHYs
    phys_chunk = b'pHYs' + struct.pack('!IIc', int(dpi[0] / 0.0254), int(dpi[1] / 0.0254), b"\x01")
    phys_chunk = struct.pack('!I', 9) + phys_chunk + struct.pack('!I', zlib.crc32(phys_chunk))

    return image_bytes[0:idat_offset] + phys_chunk + image_bytes[idat_offset:]


def _calculate_scale(old_width: int, old_height: int, new_width: int = None, new_height: int = None) -> Tuple[int, int]:
    """
    Calculate scaled image size, while keeping the aspect ratio
    :param old_height:int
    :param old_width:int
    :param new_width:int New width
    :param new_height:int New height
    :return: ndarray
    """
    if new_width and new_height:
        raise RuntimeError('Cant scale both width and height')
    if new_height:
        # noinspection PyUnresolvedReferences
        image_scale = new_height / old_height
    elif new_width:
        # noinspection PyUnresolvedReferences
        image_scale = new_width / old_width
    else:
        image_scale = 1

    return int(old_width * image_scale), int(old_height * image_scale)
